/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.zip.DataFormatException;

import morpheus_ply.*;

/**
 * This function creates a scan object from the PLY file cleans up the invalid vertices in the scan 
 * 
 * @author Detelina Stoyanova
 * @version 20171117
 * 
 */
public class BoneScan extends Thread
{
	
	// number of vertices and faces in the ply file
	private int numVertices;
	private int numFaces;
	
	
	// lists of all coordinates before cleaning them up
	private double [] initX;
	private double [] initY;
	private double [] initZ;
	
	// list of coordinates after cleaning them up
	private double [] x;
	private double [] y;
	private double [] z;
	
	// list of vertices
	private int [] v1;
	private int [] v2;
	private int [] v3;
	
	// temporary storage for cleaning up vertices that are not part of the scan
	private int [] initVertexIndex;
	private int [] vertexIndex;
	private boolean [] flag;
	
	// some ply files have a 0 vertex that is valid part of a triangle, others don't
	private int isZeroIndexValid;
	
	public BoneScan(String fileText) throws FileNotFoundException, IOException, DataFormatException
	{
		// get the filename and start working
		create(fileText);
	}
	
	private void create(String fileName) throws FileNotFoundException, IOException, DataFormatException
	{
		System.out.println(fileName);
		Ply myPly = new Ply(fileName);
		numVertices = myPly.getVertexNumber();
		numFaces = myPly.getFaceNumber();
		
		System.out.println("vertices = " + numVertices);
		System.out.println("faces = " + numFaces);
		
		v1 = new int[numFaces];
		v2 = new int[numFaces];
		v3 = new int[numFaces];
		
		int [][] tempFaces = myPly.getFaces();

		for (int i = 0; i < numFaces; i++)
		{
			v1[i] = tempFaces[i][0];
			v2[i] = tempFaces[i][1];
			v3[i] = tempFaces[i][2];
		}
		
		double [] initX1 = myPly.getX();
		double [] initY1 = myPly.getY();
		double [] initZ1= myPly.getZ();
		
		// this loops checks whether a vertex with index 0 is valid. If it is then it adds 1 to all vertices
		// so that the algorithm works the same way for files where the 0 index is valid and for those where it's not valid
		isZeroIndexValid = 0;
		for (int i = 0; i < numFaces; i++)
		{
			if (v1[i] == 0 || v2[i] == 0 || v3[i] == 0)
			{
				isZeroIndexValid = 1;
				
				for (int j = 0; j < numFaces; j++)
				{
					v1[j] ++;
					v2[j] ++;
					v3[j] ++;
				}
				
				numVertices ++;
				initX = new double [numVertices];
				initY = new double [numVertices];
				initZ = new double [numVertices];
				initX[0] = 0; initY[0] = 0; initZ[0] = 0;
				
				for (int j = 1; j < numVertices; j++)
				{
					initX[j] = initX1[j-1];
					initY[j] = initY1[j-1];
					initZ[j] = initZ1[j-1];
				}
				
				break;
			}
		}
		
		if (isZeroIndexValid == 0)
		{
			initX = initX1;
			initY = initY1;
			initZ = initZ1;
		}
		
		initVertexIndex = new int[numVertices];
		for (int i = 0; i < numVertices; i ++ )
		{
			initVertexIndex[i] = i;
		}
		
		// calls a function to check whether there are any vertices that are not part of a triangle
		selectValidVertex();
	}
	
	private void selectValidVertex()
	{	
		flag = new boolean[numVertices];
		for ( int i = 0; i < numVertices; i++)
		{
			flag[i] = false;
		}
		
		// this checks if there's an index vertex that is not part of a triangle
		for ( int i = 0; i < numVertices; i++)
   		{
   			for ( int j = 0; j < numFaces; j++)
   			{
   				if ( v1[j] == i || v2[j] == i || v3[j] == i )
   				{
   					flag[i] = true;
   					break;
   				}
   			}
   		}
		
		// this checks how many vertices are valid and initializes new storage
		int counter = 0;
		for ( int i = 0; i < numVertices; i++)
		{
			if ( flag[i] == false )
			{
				counter ++;
			}
		}
		
		// new storage initialization
		numVertices -= counter;
		
		x = new double[numVertices];
		y = new double[numVertices];
		z = new double[numVertices];
		
		vertexIndex = new int[numVertices];

		// this copies the valid vertices to the new arrays
		int count = 0;
		for ( int i = 0; i < numVertices + counter; i++)
		{
			if ( flag[i] == false )
			{
				count ++;
			}
			else
			{ 
				x[i-count] = initX[i];
				y[i-count] = initY[i];
				z[i-count] = initZ[i];
				
				vertexIndex[i-count] = initVertexIndex[i];
			}
		}

		// this updates the vertex indices
		for ( int i = 0; i < numVertices; i++ )
		{
			for ( int j = 0; j < numFaces; j++)
			{
				if (v1[j] == vertexIndex[i])
				{
					v1[j] = i+1;
				}
				if (v2[j] == vertexIndex[i])
				{
					v2[j] = i+1;
				}
				if (v3[j] == vertexIndex[i])
				{
					v3[j] = i+1;
				}
			}
		}

		for (int i = 0; i < numFaces; i++)
		{
			v1[i] --;
			v2[i] --;
			v3[i] --;
		}
	}
	
	public void flipY()
	{
		for (int i = 0; i < numVertices; i++)
		{
			y[i] = -1 * y[i];
		}
	}
	
	public int getNumVertices()
	{
		return numVertices;
	}
	
	public int getNumFaces()
	{
		return numFaces;
	}
	
	public void updateCoordinates(double [] ax, double [] ay, double [] az)
	{
		x = ax; 
		y = ay;
		z = az;
	}
	
	public void setX(double [] ax)
	{
		x = ax;
	}
	
	public void setY(double [] ay)
	{
		y = ay;
	}
	
	public void setZ(double [] az)
	{
		z = az;
	}
	
	public double[] getX()
	{
		return x;
	}
	
	public double[] getY()
	{
		return y;
	}
	
	public double[] getZ()
	{
		return z;
	}
	
	public double getX(int index)
	{
		return x[index];
	}
	
	public double getY(int index)
	{
		return y[index];
	}
	
	public double getZ(int index)
	{
		return z[index];
	}
	
	public int[] getV1()
	{
		return v1;
	}
	
	public int[] getV2()
	{
		return v2;
	}
	
	public int[] getV3()
	{
		return v3;
	}
	
	public int getV1(int index)
	{
		return v1[index];
	}
	
	public int getV2(int index)
	{
		return v2[index];
	}
	
	public int getV3(int index)
	{
		return v3[index];
	}
}
