/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;

/**
 * This function creates the meshes used by the TPS algorithm and for the calculation of the ventral curvature
 * 
 * @author Detelina Stoyanova
 * @version 20171117
 * 
 */
public class Mesh 
{
	private int numPtsX; // mesh size in x-dimension
	private int numPtsY; // mesh size in y-dimension
	
	private int initTotalPts;
	private int totalPts;
	private int numOutlinePts;

	// distance between the points
	private double dx;
	private double dy;
	
	private double[] boneScanX;
	private double[] boneScanY;
	private double[] boneScanZ;
	
	private double[] boneScanPtsX;
	private double[] boneScanPtsY;
	private double[] boneScanPtsZ;
	
	private int[] boneScanV1;
	private int[] boneScanV2;
	private int[] boneScanV3;
	
	private int boneScanNumFaces;
	
	private double boneScanMaxX, boneScanMaxY;
	private double boneScanMinX, boneScanMinY;
	private double boneScanMinZ, boneScanMaxZ;
	
	private double[] initX;
	private double[] initY;
	private double[] initZ;
	
	private double[] x;
	private double[] y;
	private double[] z;
	
	private double[] outlineX;
	private double[] outlineY;
	private double[] outlineZ;
	
	private double[] newZ;
	private boolean[] flag;
	
	public Mesh(int aNumPtsX, int af, double [] ax, double [] ay, double [] az, int [] av1, int [] av2, int [] av3, double amaxx, double amaxy, double aminx, double aminy, double amaxz, double aminz)
	{	
		numPtsX = aNumPtsX;
		boneScanNumFaces = af;
		boneScanX = ax;
		boneScanY = ay;
		boneScanZ = az;
		boneScanV1 = av1;
		boneScanV2 = av2;
		boneScanV3 = av3;
		
		boneScanMaxX = amaxx;
		boneScanMaxY = amaxy;
		boneScanMinX = aminx;
		boneScanMinY = aminy;
		boneScanMaxZ = amaxz;
		boneScanMinZ = aminz;
		
		generateInitialMesh();
		generateFinalMesh();
		generateVentralOutlineMesh();
	}
	
	private void generateInitialMesh()
	{
		dx = (boneScanMaxX - boneScanMinX) / (numPtsX - 1);

		double numptsy = (boneScanMaxY - boneScanMinY) / (dx);
		numPtsY = (int) Math.round(numptsy + 1); 
		dy = dx;

		initTotalPts = numPtsX * numPtsY;

		// initialize the arrays with mesh points
		initX = new double[initTotalPts];
		initY = new double[initTotalPts];
		initZ = new double[initTotalPts];
		
		double curLevelX = boneScanMinX;
		double curLevelY = boneScanMinY;

		for ( int i = 0; i < numPtsY; i++)
		{
			curLevelX = boneScanMinX;
			for ( int j = 0; j < numPtsX; j++)
			{
				initX[i*numPtsX + j] = curLevelX; 
				initY[i*numPtsX + j] = curLevelY;
				curLevelX += dx; 	 
			}
			curLevelY += dy;
		}

		for ( int i = 0; i < initTotalPts; i++ )
		{
			initZ[i] = 0.0;
		}
	}
	
	private void generateFinalMesh()
	{
		//double A, B, C, D;
		double [] A = new double[boneScanNumFaces];
		double [] B = new double[boneScanNumFaces];
		double [] C = new double[boneScanNumFaces];
		double [] D = new double[boneScanNumFaces];

		newZ = new double[initTotalPts];
		flag = new boolean[initTotalPts];
		
		double pqX, pqY, pqZ, prX, prY, prZ;
		double Qx, Qy, Qz, Px, Py, Pz;
		double n1x, n1y, n1z, n2x, n2y, n2z, n3x, n3y, n3z;
		double len1, len2, len3;
		
		for (int j = 0; j < boneScanNumFaces; j++)
		{
			pqX = boneScanX[boneScanV2[j]] - boneScanX[boneScanV1[j]];
			pqY = boneScanY[boneScanV2[j]] - boneScanY[boneScanV1[j]];
			pqZ = boneScanZ[boneScanV2[j]] - boneScanZ[boneScanV1[j]];
			
			prX = boneScanX[boneScanV3[j]] - boneScanX[boneScanV1[j]];
			prY = boneScanY[boneScanV3[j]] - boneScanY[boneScanV1[j]];
			prZ = boneScanZ[boneScanV3[j]] - boneScanZ[boneScanV1[j]];
			
			A[j] = pqY*prZ-pqZ*prY;
			B[j] = -(pqX*prZ-pqZ*prX);
			C[j] = pqX*prY-pqY*prX;
			
			D[j] = A[j]*(boneScanX[boneScanV1[j]]) + B[j]*(boneScanY[boneScanV1[j]]) + C[j]*(boneScanZ[boneScanV1[j]]);
		}
		
		int count = 0;
		for (int i = 0; i < initTotalPts; i++)
		{
			flag[i] = false;
			for (int j = 0; j < boneScanNumFaces; j++)
			{
				if ( boneScanX[boneScanV1[j]] > initX[i] - 2*dx && boneScanX[boneScanV1[j]] < initX[i] + 2*dx && boneScanY[boneScanV1[j]] > initY[i] - 2*dy && boneScanY[boneScanV1[j]] < initY[i] + 2*dy) 
				{
					if ( flag[i] == false )
					{			
						if ( Math.abs(C[j]) > 0.00001 )
						{
							newZ[i] = ( D[j] - A[j]*initX[i] - B[j]*initY[i] )/C[j];
					
							/*------------------------------- END MARCH ---------------------------------------*/
					
							Qx = boneScanX[boneScanV1[j]]-initX[i];
							Qy = boneScanY[boneScanV1[j]]-initY[i];
							Qz = boneScanZ[boneScanV1[j]]-newZ[i];
							
							Px = boneScanX[boneScanV2[j]]-initX[i];
							Py = boneScanY[boneScanV2[j]]-initY[i];
							Pz = boneScanZ[boneScanV2[j]]-newZ[i];

							n1x = Py*Qz - Pz*Qy;
							n1y = Pz*Qx - Px*Qz;
							n1z = Px*Qy - Py*Qx;
							len1 = Math.sqrt(n1x*n1x + n1y*n1y + n1z*n1z);
							n1x = n1x/len1;
							n1y = n1y/len1;
							n1z = n1z/len1;
							
							Qx = boneScanX[boneScanV2[j]] - initX[i];
							Qy = boneScanY[boneScanV2[j]] - initY[i];
							Qz = boneScanZ[boneScanV2[j]] - newZ[i];
							
							Px = boneScanX[boneScanV3[j]] - initX[i];
							Py = boneScanY[boneScanV3[j]] - initY[i];
							Pz = boneScanZ[boneScanV3[j]] - newZ[i];

							n2x = Py*Qz - Pz*Qy;
							n2y = Pz*Qx - Px*Qz;
							n2z = Px*Qy - Py*Qx;
							len2 = Math.sqrt(n2x*n2x + n2y*n2y + n2z*n2z);
							n2x = n2x/len2;
							n2y = n2y/len2;
							n2z = n2z/len2;
							
							Qx = boneScanX[boneScanV3[j]] - initX[i];
							Qy = boneScanY[boneScanV3[j]] - initY[i];
							Qz = boneScanZ[boneScanV3[j]] - newZ[i];
							
							Px = boneScanX[boneScanV1[j]] - initX[i];
							Py = boneScanY[boneScanV1[j]] - initY[i];
							Pz = boneScanZ[boneScanV1[j]] - newZ[i];

							n3x = Py*Qz - Pz*Qy;
							n3y = Pz*Qx - Px*Qz;
							n3z = Px*Qy - Py*Qx;
							len3 = Math.sqrt(n3x*n3x + n3y*n3y + n3z*n3z);
							n3x = n3x/len3;
							n3y = n3y/len3;
							n3z = n3z/len3;
							
							if ( (1.0 - (n1x*n2x + n1y*n2y + n1z*n2z) > -0.00001) && (1.0 - (n1x*n2x + n1y*n2y + n1z*n2z) < 0.00001) && (1.0 - (n3x*n2x + n3y*n2y + n3z*n2z)  > -0.00001) && (1.0 - (n3x*n2x + n3y*n2y + n3z*n2z)  < 0.00001) )
							{
								count ++;
								flag[i] = true;
								break;
							}
							else 
							{	
								newZ[i] = 100;
							}
						}
						else
						{
							newZ[i] = 100;
						}
					}
					else
					{
						newZ[i] = 100;
					}
				}
			}
		}

		totalPts = count;
		
		x = new double[totalPts];
		y = new double[totalPts];
		z = new double[totalPts];
		
		boneScanPtsX = new double[totalPts];
		boneScanPtsY = new double[totalPts];
		boneScanPtsZ = new double[totalPts];
		
		count = 0;
		
		for (int i = 0; i < initTotalPts; i++)
		{
			if ( flag[i] == true)
			{
				x[count] = initX[i];
				y[count] = initY[i];
				z[count] = initZ[i];
				
				boneScanPtsX[count] = initX[i];
				boneScanPtsY[count] = initY[i];
				boneScanPtsZ[count] = newZ[i];
				
				count ++;
			}	
		}
		
		for (int i = 0; i < totalPts; i++ )
		{
			if ( x[i] > boneScanMaxX || x[i] < boneScanMinX )
				x[i] = 0;
			if ( y[i] > boneScanMaxY || y[i] < boneScanMinY )
				y[i] = 0;
			if ( z[i] > boneScanMaxZ || z[i] < boneScanMinZ )
				z[i] = 0;
		}
	}
	
	private void generateVentralOutlineMesh()
	{
		numOutlinePts = 0;
		
		boolean [] isOutlinePt = new boolean[totalPts];
		for (int i = 0; i < totalPts; i++)
		{
			if ( y[i] >= 0 )
			{ 
				isOutlinePt[i] = true;
			}
			else
			{
				isOutlinePt[i] = false;
			}
		}
		
		
		for (int i = 0; i < totalPts; i++)
		{
			if ( y[i] >= 0 )
			{
				for (int j = 0; j < totalPts; j++)
				{
					if ( x[i] == x[j] && y[j] > y[i])
					{
						isOutlinePt[i] = false;
					}
				}
			}
		}
		
		for (int i = 0; i < totalPts; i++)
		{
			if ( isOutlinePt[i] )
			{
				numOutlinePts ++;
			}
		}
		
		outlineX = new double[numOutlinePts];
		outlineY = new double[numOutlinePts];
		outlineZ = new double[numOutlinePts];
		
		int counter = 0;
		for (int i = 0; i < totalPts; i++)
		{
			if ( isOutlinePt[i] )
			{
				outlineX[counter] = x[i];
				outlineY[counter] = y[i];
				outlineZ[counter] = 0;
				
				counter ++;
			}
		}
	}
	
	public int getNumOutlinePts()
	{
		return numOutlinePts;
	}
	
	public int getMeshTotalPts()
	{
		return totalPts;
	}
	
	public int getInitMeshTotalPts()
	{
		return initTotalPts;
	}
	
	public double[] getInitX()
	{
		return initX;
	}
	
	public double[] getInitY()
	{
		return initY;
	}
	
	public double[] getInitZ()
	{
		return initZ;
	}
	
	public double getInitX(int index)
	{
		return initX[index];
	}
	
	public double getInitY(int index)
	{
		return initY[index];
	}
	
	public double getInitZ(int index)
	{
		return initZ[index];
	}
	
	public double[] getX()
	{
		return x;
	}
	
	public double[] getY()
	{
		return y;
	}
	
	public double[] getZ()
	{
		return z;
	}
	
	public double getX(int index)
	{
		return x[index];
	}
	
	public double getY(int index)
	{
		return y[index];
	}
	
	public double getZ(int index)
	{
		return z[index];
	}
	
	public double[] getScanPtsX()
	{
		return boneScanPtsX;
	}
	
	public double[] getScanPtsY()
	{
		return boneScanPtsY;
	}
	
	public double[] getScanPtsZ()
	{
		return boneScanPtsZ;
	}
	
	public double getScanPtsX(int index)
	{
		return boneScanPtsX[index];
	}
	
	public double getScanPtsY(int index)
	{
		return boneScanPtsY[index];
	}
	
	public double getScanPtsZ(int index)
	{
		return boneScanPtsZ[index];
	}
	
	public double[] getOutlinePtsX()
	{
		return outlineX;
	}
	
	public double[] getOutlinePtsY()
	{
		return outlineY;
	}
	
	public double[] getOutlinePtsZ()
	{
		return outlineZ;
	}
	
	public double getOutlinePtsX(int index)
	{
		return outlineX[index];
	}
	
	public double getOutlinePtsY(int index)
	{
		return outlineY[index];
	}
	
	public double getOutlinePtsZ(int index)
	{
		return outlineZ[index];
	}
}
