/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;

import Jama.Matrix;

/**
 * This function calculates the ventral curvature of the pubic symphysis
 * 
 * @author Detelina Stoyanova
 * @version 20171117
 * 
 */

public class VentralCurvature 
{
	// the x and y coordinates of the sample points on the outline of the ventral curvature
	private double[] x;
	private double[] y;
	
	// the number of sample points
	private int numPts;
	
	private double curvatureValue;
	
	public VentralCurvature(int an, double[] ax, double[] ay)
	{
		// initialize the data
		numPts = an;
		x = ax;
		y = ay;
		
		calculateCurvature();
	}
	
	//for details of this algorithm check Stoyanova et al., 2017
	// http://onlinelibrary.wiley.com/doi/10.1111/1556-4029.13439/full
	private void calculateCurvature()
	{
		double[][] array = new double[numPts][4];
		for (int i = 0; i < numPts; i++)
		{
			array[i][0] = x[i]*x[i] + y[i]*y[i];
			array[i][1] = x[i];
			array[i][2] = y[i];
			array[i][3] = 1.0;
		}
		
		Matrix A = new Matrix(array);
		Matrix AtA = A.transpose().times(A);
		
		Matrix V = AtA.svd().getV();
		
		double [][] eigenVectors = V.getArray();
		double [] smallestEigenVector = {eigenVectors[0][3], eigenVectors[1][3], eigenVectors[2][3], eigenVectors[3][3] };

		double xc = -smallestEigenVector[1]/(2*smallestEigenVector[0]);

		double yc = -smallestEigenVector[2]/(2*smallestEigenVector[0]);

		double r = Math.sqrt(xc*xc + yc*yc - smallestEigenVector[3]/smallestEigenVector[0]);
		
		curvatureValue = 1./r;
	}

	public double getVentralCurvatureValue()
	{
		return curvatureValue;
	}
}
