/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;

import com.jogamp.opengl.util.Animator;
import javax.media.opengl.awt.GLCanvas;
import javax.media.opengl.GL2;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLProfile;
import javax.swing.JFrame;
import javax.media.opengl.glu.*;

/**
 * This function visualizes the scan and the meshes
 * 
 * @author Detelina Stoyanova
 * @version 20171117
 * 
 */
public class forAgeVis extends JFrame implements GLEventListener
{
	private static final long serialVersionUID = 1L;
	private GLU glu;
    private GLCapabilities caps;
    private GLCanvas canvas;
    
    // flags that determine whether the mesh should be rendered or not
    private boolean pcaFlag = false;
    private boolean initMeshFlag = false;
    private boolean finalMeshFlag = false;
    private boolean curvatureMeshFlag = false;
    
    // the vertices of the scan
    private float [][] pcaVertices;
    // the faces (triangles) indeces of the scan
    private int [][] pcaFaces;
    // the normals to the faces
    private float [][] pcaNormals;
    // initial mesh used for the TPS algorithm
    private float [][] initMesh;
    // the final mesh (with z-values) used for the TPS algorithm
    private float [][] finalMesh;
    // the semi-landmarks selected on the ventral margin
    private float [][] curvatureMesh;
    
    public forAgeVis()
    {	
        super("forAge 20171117 Plotter");
        
        caps = new GLCapabilities(GLProfile.getGL2ES1());
        canvas = new GLCanvas(caps);
        canvas.addGLEventListener(this);

        getContentPane().add(canvas);
    }
    
    public void setPCAVertices(float [][] pv)
    {
    	pcaVertices = pv;
    }
    
    public void setPCAFaces(int [][] pf)
    {
    	pcaFaces = pf;
    	calculateNormalVector();
    }
    
    public void setInitMesh(float [][] im)
    {
    	initMesh = im;
    }
    
    public void setFinalMesh(float [][] fm)
    {
    	finalMesh = fm;
    }
    
    public void setCurvatureMesh(float [][] cm)
    {
    	curvatureMesh = cm;
    }
    
    public void setPCAFlag(boolean arg)
    {
    	pcaFlag = arg;
    }
    
    public void setInitMeshFlag(boolean arg)
    {
    	initMeshFlag = arg;
    }
    
    public void setFinalMeshFlag(boolean arg)
    {
    	finalMeshFlag = arg;
    }
    
    public void setCurvatureMeshFlag(boolean arg)
    {
    	curvatureMeshFlag = arg;
    }
    
    public void run()
    {
        setSize(600, 600);
        setLocationRelativeTo(null);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setVisible(true);
        canvas.requestFocusInWindow();
    }

    /**
     * Initializes OpenGL properties and defines lighting.
     */
    public void init(GLAutoDrawable drawable)
    {	
        GL2 gl = (GL2) drawable.getGL();
        glu = new GLU();
        
        float ambientLight[] = { 0.3f, 0.3f, 0.3f, 1.0f };
        float specularlight[] = {1.0f, 1.0f, 1.0f, 1.0f};
        
        gl.glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
        gl.glShadeModel(GL2.GL_SMOOTH);
        gl.glEnable(GL2.GL_LIGHTING);
        
        float light_pos0[] = {2.0f, -2.0f, -5.0f, 0.0f};
        float light_pos1[] = {0.0f, 0.0f, 5.0f, 0.0f};

        gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_AMBIENT, ambientLight, 0);
        gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_POSITION, light_pos0, 0);
        gl.glEnable(GL2.GL_LIGHT0);

        gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_SPECULAR, specularlight, 0);
        gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_POSITION, light_pos1, 0);
        gl.glEnable(GL2.GL_LIGHT1);

        gl.glEnable(GL2.GL_COLOR_MATERIAL);
        gl.glEnable(GL2.GL_NORMALIZE);
        gl.glEnable(GL2.GL_DEPTH_TEST);
        
        final Animator animator = new Animator(canvas);
        animator.start();
    }

    /**
     * Displays the axis, the positioned scan, the initial rectangular mesh and the final mesh
     */
    public void display(GLAutoDrawable drawable)
    {	
        GL2 gl = (GL2) drawable.getGL();

        gl.glClear(GL2.GL_COLOR_BUFFER_BIT | GL2.GL_DEPTH_BUFFER_BIT);
        
        // x - axis
        gl.glColor3f(0.0f, 0.0f, 0.0f);
        gl.glBegin(GL2.GL_POLYGON);
	    	gl.glVertex3f(-1.0f, 0.002f, -0.1f);
	        gl.glVertex3f(1.0f, 0.002f, -0.1f);
	        
	        gl.glVertex3f(1.0f, -0.002f, -0.1f);
	        gl.glVertex3f(-1.0f, -0.002f, -0.1f);
        gl.glEnd();
        
        // y-axis
        gl.glColor3f(0.0f, 0.0f, 0.0f);
        gl.glBegin(GL2.GL_POLYGON);    
	        gl.glVertex3f( 0.002f, -1.0f, -0.1f);
	        gl.glVertex3f( 0.002f, 1.0f, -0.1f);
	        
	        gl.glVertex3f( -0.002f, 1.0f, -0.1f);
	        gl.glVertex3f( -0.002f, -1.0f, -0.1f);
        gl.glEnd();
        
        // draw sphere of the center of the coordinate system 
        GLUquadric center = glu.gluNewQuadric();
        glu.gluQuadricDrawStyle(center, GLU.GLU_FILL);
        glu.gluQuadricNormals(center, GLU.GLU_SMOOTH);
        gl.glColor3f(0.0f, 0.0f, 0.0f);
		glu.gluSphere(center, 0.001, 5, 5);
        
		// background gradient
        gl.glBegin(GL2.GL_POLYGON);
	        gl.glColor3f(0.2f, 0.3f, 1.0f);
	    	gl.glVertex3f(-0.6f, -0.8f, -0.2f);
	        gl.glVertex3f(0.6f, -0.8f, -0.2f);

	        gl.glColor3f(0.6f, 0.8f, 1.0f);
	        gl.glVertex3f(0.6f, 0.55f, -0.2f);
	        gl.glVertex3f(-0.6f, 0.55f, -0.2f);
        gl.glEnd();
        
        // display the positioned scan if menu option is selected
        if ( pcaFlag == true )
        {
        	gl.glBegin(GL2.GL_TRIANGLES);
        	for(int i = 0; i < pcaFaces.length; i++)
        	{
        		gl.glNormal3f(pcaNormals[pcaFaces[i][0]][0], pcaNormals[pcaFaces[i][0]][1], pcaNormals[pcaFaces[i][0]][2]);
        		gl.glColor3f(0.5f, 0.5f, 0.5f);
        		gl.glVertex3f(pcaVertices[pcaFaces[i][0]][0]/10, pcaVertices[pcaFaces[i][0]][1]/10, pcaVertices[pcaFaces[i][0]][2]/10);
        		
        		gl.glNormal3f(pcaNormals[pcaFaces[i][1]][0], pcaNormals[pcaFaces[i][1]][1], pcaNormals[pcaFaces[i][1]][2]);
        		gl.glColor3f(0.5f, 0.5f, 0.5f);
        		gl.glVertex3f(pcaVertices[pcaFaces[i][1]][0]/10, pcaVertices[pcaFaces[i][1]][1]/10, pcaVertices[pcaFaces[i][1]][2]/10);
        		
        		gl.glNormal3f(pcaNormals[pcaFaces[i][2]][0], pcaNormals[pcaFaces[i][2]][1], pcaNormals[pcaFaces[i][2]][2]);
        		gl.glColor3f(0.5f, 0.5f, 0.5f);
        		gl.glVertex3f(pcaVertices[pcaFaces[i][2]][0]/10, pcaVertices[pcaFaces[i][2]][1]/10, pcaVertices[pcaFaces[i][2]][2]/10);
        	}
        	gl.glEnd();
        }
        
        // display the initial rectangular mesh if menu option is selected
        if ( initMeshFlag == true )
        {
        	GLUquadric shape = glu.gluNewQuadric();
            glu.gluQuadricDrawStyle(shape, GLU.GLU_FILL);
            glu.gluQuadricNormals(shape, GLU.GLU_SMOOTH);
        	for(int i = 0; i < initMesh.length; i++)
        	{
        		gl.glColor3f(0.0f, 0f, 0.0f);
        		gl.glPushMatrix();
        		gl.glTranslatef(initMesh[i][0]/10, initMesh[i][1]/10, initMesh[i][2]/10);
        		glu.gluSphere(shape, 0.006, 10, 10);
        		gl.glPopMatrix();
        	}
        }
        
        // display the final mesh if menu option is selected
        if ( finalMeshFlag == true )
        {
        	GLUquadric shape = glu.gluNewQuadric();
            glu.gluQuadricDrawStyle(shape, GLU.GLU_FILL);
            glu.gluQuadricNormals(shape, GLU.GLU_SMOOTH);
        	for(int i = 0; i < finalMesh.length; i++)
        	{
        		gl.glColor3f(0f, 0f, 0f);
        		gl.glPushMatrix();
        		gl.glTranslatef(finalMesh[i][0]/10, finalMesh[i][1]/10, finalMesh[i][2]/10+0.1f);
        		glu.gluSphere(shape, 0.006, 10, 10);
        		gl.glPopMatrix();
        	}
        }
        
        // display the curvature mesh if menu option is selected
        if ( curvatureMeshFlag == true )
        {
        	GLUquadric shape = glu.gluNewQuadric();
            glu.gluQuadricDrawStyle(shape, GLU.GLU_FILL);
            glu.gluQuadricNormals(shape, GLU.GLU_SMOOTH);
        	for(int i = 0; i < curvatureMesh.length; i++)
        	{
        		gl.glColor3f(0f, 0f, 0f);
        		gl.glPushMatrix();
        		gl.glTranslatef(curvatureMesh[i][0]/10, curvatureMesh[i][1]/10, curvatureMesh[i][2]/10+0.1f);
        		glu.gluSphere(shape, 0.006, 10, 10);
        		gl.glPopMatrix();
        	}
        }
        
        gl.glFlush();
    }

    /**
     * Sets the View and Projection matrices and handles resizing of the window. 
     */
    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height)
    {
        GL2 gl = (GL2) drawable.getGL();
        GLU glu = new GLU();
        
        gl.glViewport(0, 0, width, height);
        gl.glMatrixMode(GL2.GL_PROJECTION);
        gl.glLoadIdentity();
        glu.gluPerspective(10.0, (float)width/(float)height, 2.0, 10.0);   
        gl.glMatrixMode(GL2.GL_MODELVIEW);
        gl.glLoadIdentity();
        glu.gluLookAt(0.0, 0.0, 6.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    }

    /**
     * Calculates the normal vectors to the faces of the 3D bones scan using its vetrex x, y, z coordinates
     */
    private void calculateNormalVector()
    {
    	pcaNormals = new float[pcaVertices.length][3];

    	int [] counter_x = new int[pcaVertices.length];
    	int [] counter_y = new int[pcaVertices.length];
    	int [] counter_z = new int[pcaVertices.length];

        for(int i = 0; i < pcaVertices.length; i++)
    	{
        	pcaNormals[i][0] = 0;
        	pcaNormals[i][1] = 0;
        	pcaNormals[i][2] = 0;

            counter_x[i] = 0;
            counter_y[i] = 0;
            counter_z[i] = 0;
    	}
    	
        float Qx, Qy, Qz, Px, Py, Pz;
        
    	for(int i = 0; i < pcaFaces.length; i++)
    	{ 
            Qx = pcaVertices[pcaFaces[i][1]][0] - pcaVertices[pcaFaces[i][0]][0]; 
            Qy = pcaVertices[pcaFaces[i][1]][1] - pcaVertices[pcaFaces[i][0]][1]; 
            Qz = pcaVertices[pcaFaces[i][1]][2] - pcaVertices[pcaFaces[i][0]][2]; 
            Px = pcaVertices[pcaFaces[i][2]][0] - pcaVertices[pcaFaces[i][0]][0]; 
            Py = pcaVertices[pcaFaces[i][2]][1] - pcaVertices[pcaFaces[i][0]][1]; 
            Pz = pcaVertices[pcaFaces[i][2]][2] - pcaVertices[pcaFaces[i][0]][2];    
            
            pcaNormals[pcaFaces[i][0]][0] += Py*Qz - Pz*Qy;
            pcaNormals[pcaFaces[i][0]][1] += Pz*Qx - Px*Qz;
            pcaNormals[pcaFaces[i][0]][2] += Px*Qy - Py*Qx;
            
            counter_x[pcaFaces[i][0]]++;
            counter_y[pcaFaces[i][0]]++;
            counter_z[pcaFaces[i][0]]++;
            
            pcaNormals[pcaFaces[i][1]][0] += Py*Qz - Pz*Qy;
            pcaNormals[pcaFaces[i][1]][1] += Pz*Qx - Px*Qz;
            pcaNormals[pcaFaces[i][1]][2] += Px*Qy - Py*Qx;

            counter_x[pcaFaces[i][1]]++;
            counter_y[pcaFaces[i][1]]++;
            counter_z[pcaFaces[i][1]]++;

            pcaNormals[pcaFaces[i][2]][0] += Py*Qz - Pz*Qy;
            pcaNormals[pcaFaces[i][2]][1] += Pz*Qx - Px*Qz;
            pcaNormals[pcaFaces[i][2]][2] += Px*Qy - Py*Qx;

            counter_x[pcaFaces[i][2]]++;
            counter_y[pcaFaces[i][2]]++;
            counter_z[pcaFaces[i][2]]++;  
    	}

    	for(int i = 0; i < pcaVertices.length; i++)
    	{
    		pcaNormals[i][0] = pcaNormals[i][0]/counter_x[i];
    		pcaNormals[i][1] = pcaNormals[i][1]/counter_y[i];
    		pcaNormals[i][2] = pcaNormals[i][2]/counter_z[i];	
    	}
    	
    	if ( pcaNormals[10][2] > 0 )
    	{
    		for(int i = 0; i < pcaVertices.length; i++)
    		{
    			pcaNormals[i][0] = -pcaNormals[i][0];
    			pcaNormals[i][1] = -pcaNormals[i][1];
    			pcaNormals[i][2] = -pcaNormals[i][2];
    		}
    	}
    	else	
    	{
    		for(int i = 0; i < pcaVertices.length; i++)
    		{
    			pcaVertices[i][2] = -pcaVertices[i][2];
    		}
    	}
    }
    
    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
        boolean deviceChanged)
    {
    }

	@Override
	public void dispose(GLAutoDrawable arg0) {
		// TODO Auto-generated method stub
		
	}
}
